@@ -5,6 +5,7 @@ use rand_distr::Bernoulli;
 use rayon::prelude::*;
 use serde_derive::{Deserialize, Serialize};
 use std::borrow::Cow;
+use std::sync::Arc;
 
 use crate::output_covariance::OutputCovariance;
 use crate::utils::{standard_noise, standard_noise_matrix, Mask};
@@ -52,19 +53,82 @@ impl MaskedSample {
     }
 }
 
+#[derive(Debug, Clone)]
 pub struct Dataset {
-    pub(crate) data: Vec<MaskedSample>,
+    pub(crate) data: Arc<Vec<MaskedSample>>,
+    pub(crate) weights: Vec<f64>,
 }
 
 impl From<Vec<MaskedSample>> for Dataset {
     fn from(value: Vec<MaskedSample>) -> Self {
-        Dataset { data: value }
+        Dataset {
+            weights: vec![1.0; value.len()],
+            data: Arc::new(value),
+        }
+    }
+}
+
+impl FromIterator<MaskedSample> for Dataset {
+    fn from_iter<T>(iter: T) -> Self
+    where
+        T: IntoIterator<Item = MaskedSample>,
+    {
+        let data: Vec<_> = iter.into_iter().collect();
+        Self::new(data)
+    }
+}
+
+impl FromIterator<(MaskedSample, f64)> for Dataset {
+    fn from_iter<T>(iter: T) -> Self
+    where
+        T: IntoIterator<Item = (MaskedSample, f64)>,
+    {
+        let (data, weights): (Vec<_>, Vec<_>) = iter.into_iter().unzip();
+        Self::new_with_weights(data, weights)
+    }
+}
+
+impl FromParallelIterator<MaskedSample> for Dataset {
+    fn from_par_iter<T>(iter: T) -> Self
+    where
+        T: IntoParallelIterator<Item = MaskedSample>,
+    {
+        let data: Vec<_> = iter.into_par_iter().collect();
+        Self::new(data)
+    }
+}
+
+impl FromParallelIterator<(MaskedSample, f64)> for Dataset {
+    fn from_par_iter<T>(iter: T) -> Self
+    where
+        T: IntoParallelIterator<Item = (MaskedSample, f64)>,
+    {
+        let (data, weights): (Vec<_>, Vec<_>) = iter.into_par_iter().unzip();
+        Self::new_with_weights(data, weights)
     }
 }
 
 impl Dataset {
     pub fn new(data: Vec<MaskedSample>) -> Dataset {
-        Dataset { data }
+        Dataset {
+            weights: vec![1.0; data.len()],
+            data: Arc::new(data),
+        }
+    }
+
+    pub fn new_with_weights(data: Vec<MaskedSample>, weights: Vec<f64>) -> Dataset {
+        assert_eq!(data.len(), weights.len());
+        Dataset {
+            data: Arc::new(data),
+            weights,
+        }
+    }
+
+    pub fn with_weights(&self, weights: Vec<f64>) -> Dataset {
+        Dataset {
+            data: self.data.clone(),
+            weights,
+        }
     }
 
     pub fn len(&self) -> usize {
@@ -114,17 +178,22 @@ impl Dataset {
 pub struct PPCAModel {
     output_covariance: OutputCovariance<'static>,
     mean: DVector<f64>,
+    /// A factor to smooth out the `transform` matrix when there is little data for a
+    /// given dimension.
+    #[serde(default)]
+    smoothing_factor: f64,
 }
 
 impl PPCAModel {
-    pub fn new(isotropic_noise: f64, transform: DMatrix<f64>, mean: DVector<f64>) -> PPCAModel {
+    pub fn new(isotropic_noise: f64, transform: DMatrix<f64>, mean: DVector<f64>, smoothing_factor: f64) -> PPCAModel {
         PPCAModel {
             output_covariance: OutputCovariance::new_owned(isotropic_noise, transform),
             mean,
+            smoothing_factor,
         }
     }
 
-    pub fn init(state_size: usize, dataset: &Dataset) -> PPCAModel {
+    pub fn init(state_size: usize, dataset: &Dataset, smoothing_factor: f64) -> PPCAModel {
         assert!(!dataset.is_empty());
         let output_size = dataset.output_size().expect("dataset is not empty");
         let empty_dimensions = dataset.empty_dimensions();
@@ -142,6 +211,7 @@ impl PPCAModel {
                 transform: Cow::Owned(rand_transform),
             },
             mean: DVector::zeros(output_size),
+            smoothing_factor,
         }
     }
 
@@ -172,89 +242,125 @@ impl PPCAModel {
         1 + self.state_size() * self.output_size() + self.mean.nrows()
     }
 
+    pub fn singular_values(&self) -> DVector<f64> {
+        self.output_covariance
+            .transform
+            .column_iter()
+            .map(|column| column.norm().sqrt())
+            .collect::<Vec<_>>()
+            .into()
+    }
+
+    pub(crate) fn llk_one(&self, sample: &MaskedSample) -> f64 {
+        let sample = if !sample.is_empty() {
+            sample
+        } else {
+            return 0.0;
+        };
+
+        let sub_sample = sample.mask.mask(&(sample.data_vector() - &self.mean));
+        let sub_covariance = self.output_covariance.masked(&sample.mask);
+
+        let llk = -sub_covariance.quadratic_form(&sub_sample) / 2.0
+            - sub_covariance.covariance_log_det() / 2.0
+            - LN_2PI / 2.0 * sub_covariance.output_size() as f64;
+
+        llk
+    }
+
     pub fn llk(&self, dataset: &Dataset) -> f64 {
         dataset
             .data
             .par_iter()
-            .filter(|sample| !sample.is_empty())
-            .map(|sample| {
-                let sub_sample = sample.mask.mask(&(sample.data_vector() - &self.mean));
-                let sub_covariance = self.output_covariance.masked(&sample.mask);
-
-                -sub_covariance.quadratic_form(&sub_sample) / 2.0
-                    - sub_covariance.covariance_log_det() / 2.0
-                    - LN_2PI / 2.0 * sub_covariance.output_size() as f64
-            })
+            .zip(&dataset.weights)
+            .map(|(sample, weight)| self.llk_one(sample) * weight)
             .sum()
     }
 
+    pub fn llks(&self, dataset: &Dataset) -> DVector<f64> {
+        dataset
+            .data
+            .par_iter()
+            .map(|sample| self.llk_one(sample))
+            .collect::<Vec<_>>()
+            .into()
+    }
+
+    pub(crate) fn sample_one(&self, mask_prob: f64) -> MaskedSample {
+        let sampled_state: DVector<f64> =
+            &*self.output_covariance.transform * standard_noise(self.state_size()) + &self.mean;
+        let noise: DVector<f64> =
+            self.output_covariance.isotropic_noise * standard_noise(self.output_size());
+        let mask = Mask(
+            Bernoulli::new(1.0 - mask_prob as f64)
+                .expect("invalid mask probability")
+                .sample_iter(rand::thread_rng())
+                .take(self.output_size())
+                .collect::<BitVec>(),
+        );
+
+        MaskedSample {
+            data: mask.fillna(&(sampled_state + noise)),
+            mask,
+        }
+    }
+
     pub fn sample(&self, dataset_size: usize, mask_prob: f64) -> Dataset {
         (0..dataset_size)
             .into_par_iter()
-            .map(|_| {
-                let sampled_state: DVector<f64> = &*self.output_covariance.transform
-                    * standard_noise(self.state_size())
-                    + &self.mean;
-                let noise: DVector<f64> =
-                    self.output_covariance.isotropic_noise * standard_noise(self.output_size());
-                let mask = Mask(
-                    Bernoulli::new(1.0 - mask_prob as f64)
-                        .expect("invalid mask probability")
-                        .sample_iter(rand::thread_rng())
-                        .take(self.output_size())
-                        .collect::<BitVec>(),
-                );
+            .map(|_| self.sample_one(mask_prob))
+            .collect()
+    }
 
-                MaskedSample {
-                    data: mask.fillna(&(sampled_state + noise)),
-                    mask,
-                }
-            })
-            .collect::<Vec<_>>()
-            .into()
+    pub(crate) fn infer_one(&self, sample: &MaskedSample) -> InferredMasked {
+        if sample.is_empty() {
+            return self.uninferred();
+        }
+
+        let sub_sample = sample.mask.mask(&(sample.data_vector() - &self.mean));
+        let sub_covariance = self.output_covariance.masked(&sample.mask);
+
+        InferredMasked {
+            state: sub_covariance.estimator_transform() * sub_sample,
+            covariance: sub_covariance.estimator_covariance(),
+        }
     }
 
     pub fn infer(&self, dataset: &Dataset) -> Vec<InferredMasked> {
         dataset
             .data
             .par_iter()
-            .map(|sample| {
-                if sample.is_empty() {
-                    return self.uninferred();
-                }
-
-                let sub_sample = sample.mask.mask(&(sample.data_vector() - &self.mean));
-                let sub_covariance = self.output_covariance.masked(&sample.mask);
-
-                InferredMasked {
-                    state: sub_covariance.estimator_transform() * sub_sample,
-                    covariance: sub_covariance.estimator_covariance(),
-                }
-            })
+            .map(|sample| self.infer_one(sample))
             .collect()
     }
 
-    pub fn filter_extrapolate(&self, dataset: &Dataset) -> Dataset {
-        self.infer(dataset)
-            .into_par_iter()
-            .map(|inferred| &*self.output_covariance.transform * inferred.state())
-            .map(MaskedSample::unmasked)
-            .collect::<Vec<_>>()
-            .into()
+    pub(crate) fn smooth_one(&self, sample: &MaskedSample) -> MaskedSample {
+        MaskedSample::unmasked(self.infer_one(sample).smoothed(&self))
+    }
+
+    pub fn smooth(&self, dataset: &Dataset) -> Dataset {
+        dataset
+            .data
+            .par_iter()
+            .zip(&dataset.weights)
+            .map(|(sample, &weight)| (self.smooth_one(sample), weight))
+            .collect()
+    }
+
+    pub(crate) fn extrapolate_one(&self, sample: &MaskedSample) -> MaskedSample {
+        MaskedSample::unmasked(self.infer_one(sample).extrapolated(self, sample))
     }
 
     pub fn extrapolate(&self, dataset: &Dataset) -> Dataset {
-        self.infer(dataset)
-            .into_par_iter()
-            .zip(&dataset.data)
-            .map(|(inferred, sample)| {
-                let filtered = &*self.output_covariance.transform * inferred.state();
-                MaskedSample::unmasked(sample.mask.choose(&sample.data_vector(), &filtered))
-            })
-            .collect::<Vec<_>>()
-            .into()
+        dataset
+            .data
+            .par_iter()
+            .zip(&dataset.weights)
+            .map(|(sample, &weight)| (self.extrapolate_one(sample), weight))
+            .collect()
     }
 
+    #[must_use]
     pub fn iterate(&self, dataset: &Dataset) -> PPCAModel {
         let inferred = self.infer(dataset);
 
@@ -262,10 +368,11 @@ impl PPCAModel {
         let total_cross_moment = dataset
             .data
             .par_iter()
+            .zip(&dataset.weights)
             .zip(&inferred)
-            .map(|(sample, inferred)| {
+            .map(|((sample, &weight), inferred)| {
                 let centered_filled = sample.mask.fillna(&(sample.data_vector() - &self.mean));
-                centered_filled * inferred.state.transpose()
+                weight * centered_filled * inferred.state.transpose()
             })
             .reduce(
                 || DMatrix::zeros(self.output_size(), self.state_size()),
@@ -277,13 +384,14 @@ impl PPCAModel {
                 let total_second_moment = dataset
                     .data
                     .iter()
+                    .zip(&dataset.weights)
                     .zip(&inferred)
-                    .filter(|(sample, _)| sample.mask.0[idx])
-                    .map(|(_, inferred)| inferred.second_moment())
+                    .filter(|((sample, _), _)| sample.mask.0[idx])
+                    .map(|((_, &weight), inferred)| weight * inferred.second_moment())
                     // In case we get an empty dimension...
                     .chain([DMatrix::zeros(self.state_size(), self.state_size())])
                     .sum::<DMatrix<f64>>()
-                    + 1.0 * DMatrix::<f64>::identity(self.state_size(), self.state_size());
+                    + self.smoothing_factor * DMatrix::<f64>::identity(self.state_size(), self.state_size());
                 let cross_moment_row = total_cross_moment.row(idx).transpose();
                 total_second_moment
                     .qr()
@@ -302,10 +410,10 @@ impl PPCAModel {
         let new_transform = DMatrix::from_rows(&new_transform_rows);
 
         // Updated isotropic noise:
-        let (square_error, deviations_square_sum, total_deviation, totals) = dataset.data.par_iter().zip(&inferred)
-        .filter(|(sample, _)| !sample.is_empty())
+        let (square_error, deviations_square_sum, total_deviation, totals) = dataset.data.par_iter().zip(&dataset.weights).zip(&inferred)
+        .filter(|((sample, _), _)| !sample.is_empty())
         .map(
-            |(sample, inferred)| {
+            |((sample, &weight), inferred)| {
             let sub_covariance = self.output_covariance.masked(&sample.mask);
             let sub_transform = &*sub_covariance.transform;
             let deviation = sample.mask.fillna(
@@ -315,10 +423,10 @@ impl PPCAModel {
             );
 
             (
-                (sub_transform * &inferred.covariance).dot(&sub_transform),
-                deviation.norm_squared(),
-                deviation,
-                sample.mask.as_vector()
+                weight * (sub_transform * &inferred.covariance).dot(&sub_transform),
+                weight * deviation.norm_squared(),
+                weight * deviation,
+                weight * sample.mask.as_vector(),
             )
         }).reduce_with(|
             (square_error, deviation_square_sum, total_deviation, totals),
@@ -342,6 +450,7 @@ impl PPCAModel {
                 transform: Cow::Owned(new_transform),
                 isotropic_noise: average_square_error.sqrt(),
             },
+            smoothing_factor: self.smoothing_factor,
             mean: new_mean,
         }
     }
@@ -365,6 +474,7 @@ impl PPCAModel {
                 self.output_covariance.isotropic_noise,
                 new_transform,
             ),
+            smoothing_factor: self.smoothing_factor,
             mean: self.mean.clone(),
         }
     }
@@ -377,7 +487,7 @@ pub struct InferredMasked {
 }
 
 impl InferredMasked {
-    fn second_moment(&self) -> DMatrix<f64> {
+    pub(crate) fn second_moment(&self) -> DMatrix<f64> {
         &self.state * self.state.transpose() + &self.covariance
     }
 
@@ -385,8 +495,112 @@ impl InferredMasked {
         &self.state
     }
 
-    pub fn covariace(&self) -> &DVector<f64> {
-        &self.state
+    pub fn covariance(&self) -> &DMatrix<f64> {
+        &self.covariance
+    }
+
+    pub fn smoothed(&self, ppca: &PPCAModel) -> DVector<f64> {
+        &*ppca.output_covariance.transform * self.state() + &ppca.mean
+    }
+
+    pub fn extrapolated(&self, ppca: &PPCAModel, sample: &MaskedSample) -> DVector<f64> {
+        let filtered = self.smoothed(&ppca);
+        sample.mask.choose(&sample.data_vector(), &filtered)
+    }
+
+    /// Afraid of the big, fat matrix? The method `output_covariance_diagonal` might just
+    /// save your life.
+    pub fn smoothed_covariance(&self, ppca: &PPCAModel) -> DMatrix<f64> {
+        let covariance = &ppca.output_covariance;
+
+        DMatrix::identity(covariance.output_size(), covariance.output_size())
+            * covariance.isotropic_noise.powi(2)
+            + &*covariance.transform * &self.covariance * covariance.transform.transpose()
+    }
+
+    /// Use this not to get lost with big matrices in the output, losing CPU, memory and hair.
+    pub fn smoothed_covariance_diagonal(&self, ppca: &PPCAModel) -> DVector<f64> {
+        // Here, we will calculate `I sigma^2 + C Sxx C^T` for the unobserved samples in a
+        // clever way...
+
+        let covariance = &ppca.output_covariance;
+
+        // The `inner_inverse` part.
+        let transformed_state_covariance = &*covariance.transform * &self.covariance;
+
+        // Now comes the trick! Calculate only the diagonal elements of the
+        // `transformed_state_covariance * C^T` part.
+        let noiseless_output_diagonal = transformed_state_covariance
+            .row_iter()
+            .zip(covariance.transform.row_iter())
+            .map(|(row_left, row_right)| row_left.dot(&row_right));
+
+        // Finally, add the isotropic noise term...
+        noiseless_output_diagonal
+            .map(|noiseless_output_variance| {
+                noiseless_output_variance + covariance.isotropic_noise.powi(2)
+            })
+            .collect::<Vec<_>>()
+            .into()
+    }
+
+    /// Afraid of the big, fat matrix? The method `output_covariance_diagonal` might just
+    /// save your life.
+    pub fn extrapolated_covariance(&self, ppca: &PPCAModel, sample: &MaskedSample) -> DMatrix<f64> {
+        let negative = sample.mask().negate();
+
+        if !negative.0.any() {
+            return DMatrix::zeros(ppca.output_size(), ppca.output_size());
+        }
+
+        let sub_covariance = ppca.output_covariance.masked(&negative);
+
+        let output_covariance =
+            DMatrix::identity(sub_covariance.output_size(), sub_covariance.output_size())
+                * sub_covariance.isotropic_noise.powi(2)
+                + &*sub_covariance.transform
+                    * &self.covariance
+                    * sub_covariance.transform.transpose();
+
+        negative.expand_matrix(output_covariance)
+    }
+
+    /// Use this not to get lost with big matrices in the output, losing CPU, memory and hair.
+    pub fn extrapolated_covariance_diagonal(
+        &self,
+        ppca: &PPCAModel,
+        sample: &MaskedSample,
+    ) -> DVector<f64> {
+        // Here, we will calculate `I sigma^2 + C Sxx C^T` for the unobserved samples in a
+        // clever way...
+
+        let negative = sample.mask().negate();
+
+        if !negative.0.any() {
+            return DVector::zeros(ppca.output_size());
+        }
+
+        let sub_covariance = ppca.output_covariance.masked(&negative);
+
+        // The `inner_inverse` part.
+        let transformed_state_covariance = &*sub_covariance.transform * &self.covariance;
+
+        // Now comes the trick! Calculate only the diagonal elements of the
+        // `transformed_state_covariance * C^T` part.
+        let noiseless_output_diagonal = transformed_state_covariance
+            .row_iter()
+            .zip(sub_covariance.transform.row_iter())
+            .map(|(row_left, row_right)| row_left.dot(&row_right));
+
+        // Finally, add the isotropic noise term...
+        let diagonal_reduced = noiseless_output_diagonal
+            .map(|noiseless_output_variance| {
+                noiseless_output_variance + sub_covariance.isotropic_noise.powi(2)
+            })
+            .collect::<Vec<_>>()
+            .into();
+
+        negative.expand(&diagonal_reduced)
     }
 }
 
